@@ -1,3 +1,9 @@
+/* Name of this package (needed by automake) */
+#undef PACKAGE
+
+/* Version of this package (needed by automake) */
+#undef VERSION
+
 /* Define to the name of the SCCS `get' command.  */
 #undef SCCS_GET
 
@@ -12,6 +18,3 @@
 
 /* Define this if you have the `union wait' type in <sys/wait.h>.  */
 #undef HAVE_UNION_WAIT
-
-/* Define this if the POSIX.1 call `sysconf (_SC_OPEN_MAX)' works properly.  */
-#undef HAVE_SYSCONF_OPEN_MAX
