@@ -1,5 +1,5 @@
 /* Interface to `ar' archives for GNU Make.
-Copyright (C) 1988, 1989, 1990, 1991, 1992, 1993 Free Software Foundation, Inc.
+Copyright (C) 1988,89,90,91,92,93,97 Free Software Foundation, Inc.
 This file is part of GNU Make.
 
 GNU Make is free software; you can redistribute it and/or modify
@@ -20,15 +20,16 @@ the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  */
 
 #ifndef	NO_ARCHIVES
 
-#include "file.h"
+#include "filedef.h"
 #include "dep.h"
 #include <fnmatch.h>
 
 /* Defined in arscan.c.  */
-extern long int ar_scan ();
-extern int ar_member_touch ();
-extern int ar_name_equal ();
-
+extern long int ar_scan PARAMS ((char *archive, long int (*function) (), long int arg));
+extern int ar_name_equal PARAMS ((char *name, char *mem, int truncated));
+#ifndef VMS
+extern int ar_member_touch PARAMS ((char *arname, char *memname));
+#endif
 
 /* Return nonzero if NAME is an archive-member reference, zero if not.
    An archive-member reference is a name like `lib(member)'.
@@ -40,7 +41,7 @@ ar_name (name)
      char *name;
 {
   char *p = index (name, '('), *end = name + strlen (name) - 1;
-  
+
   if (p == 0 || p == name || *end != ')')
     return 0;
 
@@ -66,9 +67,10 @@ ar_parse_name (name, arname_p, memname_p)
 
   if (memname_p != 0)
     *memname_p = savestring (p + 1, end - (p + 1));
-}  
+}
 
-static long int ar_member_date_1 ();
+static long int ar_member_date_1 PARAMS ((int desc, char *mem, int truncated, long int hdrpos,
+	long int datapos, long int size, long int date, int uid, int gid, int mode, char *name));
 
 /* Return the modtime of NAME.  */
 
@@ -130,6 +132,15 @@ ar_member_date_1 (desc, mem, truncated,
 
 /* Set the archive-member NAME's modtime to now.  */
 
+#ifdef VMS
+int
+ar_touch (name)
+     char *name;
+{
+  error ("touch archive member is not available on VMS");
+  return -1;
+}
+#else
 int
 ar_touch (name)
      char *name;
@@ -182,6 +193,7 @@ ar_touch (name)
 
   return val;
 }
+#endif /* !VMS */
 
 /* State of an `ar_glob' run, passed to `ar_glob_match'.  */
 
