@@ -1,5 +1,5 @@
 /* Basic dependency engine for GNU Make.
-Copyright (C) 1988,89,90,91,92,93,94,95,96 Free Software Foundation, Inc.
+Copyright (C) 1988,89,90,91,92,93,94,95,96,97 Free Software Foundation, Inc.
 This file is part of GNU Make.
 
 GNU Make is free software; you can redistribute it and/or modify
@@ -29,10 +29,14 @@ the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  */
 #include <sys/file.h>
 #endif
 
+#ifdef  __MSDOS__
+#include "variable.h"
+#endif
+
 #ifdef VMS
 #include <starlet.h>
 #endif
-#ifdef WIN32
+#ifdef WINDOWS32
 #include <io.h>
 #endif
 
@@ -52,10 +56,6 @@ static int library_search PARAMS ((char **lib, time_t *mtime_ptr));
 
 extern time_t f_mtime PARAMS ((struct file *file, int search));
 
-#ifdef VMS
-extern int vms_stat PARAMS ((char *name, struct stat *buf));
-#endif
-
 
 /* Remake all the goals in the `struct dep' chain GOALS.  Return -1 if nothing
    was done, 0 if all goals were updated successfully, or 1 if a goal failed.
@@ -566,13 +566,46 @@ update_file_1 (file, depth)
 
   if (!must_make)
     {
-      DEBUGPR ("No need to remake target `%s'.\n");
+      if (debug_flag)
+        {
+          print_spaces(depth);
+          printf("No need to remake target `%s'", file->name);
+          if (!streq(file->name, file->hname))
+              printf("; using VPATH name `%s'", file->hname);
+          printf(".\n");
+          fflush(stdout);
+        }
+
       notice_finished_file (file);
+
+      /* Since we don't need to remake the file, convert it to use the
+         VPATH filename if we found one.  hfile will be either the
+         local name if no VPATH or the VPATH name if one was found.  */
+
+      while (file)
+        {
+          file->name = file->hname;
+          file = file->prev;
+        }
+
       return 0;
     }
 
   DEBUGPR ("Must remake target `%s'.\n");
 
+  /* It needs to be remade.  If it's VPATH and not reset via GPATH, toss the
+     VPATH */
+  if (!streq(file->name, file->hname))
+    {
+      if (debug_flag)
+        {
+          print_spaces (depth);
+          printf("  Ignoring VPATH name `%s'.\n", file->hname);
+          fflush(stdout);
+        }
+      file->ignore_vpath = 1;
+    }
+
   /* Now, take appropriate actions to remake the file.  */
   remake_file (file);
 
@@ -745,7 +778,7 @@ check_dep (file, depth, this_mtime, must_make_ptr)
       check_renamed (file);
       mtime = file_mtime (file);
       check_renamed (file);
-      if (mtime > this_mtime)
+      if (mtime != (time_t) -1 && mtime > this_mtime)
 	*must_make_ptr = 1;
 	  /* Otherwise, update all non-intermediate files we depend on,
 	     if necessary, and see whether any of them is more
@@ -1003,7 +1036,7 @@ f_mtime (file, search)
     {
       mtime = name_mtime (file->name);
 
-      if (mtime == (time_t) -1 && search)
+      if (mtime == (time_t) -1 && search && !file->ignore_vpath)
 	{
 	  /* If name_mtime failed, search VPATH.  */
 	  char *name = file->name;
@@ -1016,9 +1049,20 @@ f_mtime (file, search)
 		/* vpath_search and library_search store zero in MTIME
 		   if they didn't need to do a stat call for their work.  */
 		file->last_mtime = mtime;
-	      rename_file (file, name);
+
+              /* If we found it in VPATH, see if it's in GPATH too; if so,
+                 change the name right now; if not, defer until after the
+                 dependencies are updated. */
+              if (gpath_search (name, strlen(name) - strlen(file->name) - 1))
+                {
+                  rename_file (file, name);
+                  check_renamed (file);
+                  return file_mtime (file);
+                }
+
+	      rehash_file (file, name);
 	      check_renamed (file);
-	      return file_mtime (file);
+	      mtime = name_mtime (name);
 	    }
 	}
     }
@@ -1026,29 +1070,39 @@ f_mtime (file, search)
   {
     /* Files can have bogus timestamps that nothing newly made will be
        "newer" than.  Updating their dependents could just result in loops.
-       So notify the user of the anomaly by treating future files as
-       unrecoverably absent.  */
+       So notify the user of the anomaly with a warning.
 
-    static time_t now;
-    if (mtime != -1 && mtime > now && ! file->updated)
+       We only need to do this once, for now. */
+
+    static time_t now = 0;
+    if (!clock_skew_detected
+        && mtime != (time_t)-1 && mtime > now
+        && !file->updated)
       {
 	/* This file's time appears to be in the future.
 	   Update our concept of the present, and compare again.  */
-#ifdef VMS
-	/* Handle vms 64bit to 32bit time hack introduced in vms_stat ... */
-	static unsigned long vms_now[2]; /* assumes 32 bit long ! */
-	sys$gettim (vms_now);
-	now = ((vms_now[0]>>24) & 0xff) + ((vms_now[1]<<8) & 0xffffff00);
-	if (mtime > now)
-#else
+
 	extern time_t time ();
-	if (mtime > time (&now))
+	time (&now);
+
+#ifdef WINDOWS32
+	/*
+	 * FAT filesystems round time to nearest even second(!). Just
+	 * allow for any file (NTFS or FAT) to perhaps suffer from this
+	 * braindamage.
+	 *
+	 * Apparently, this doesn't happen with the MS-DOS/DJGPP port,
+	 * although MS-DOS and MS-Windows 3.X/9X also use FAT filesystems.
+	 */
+	if (mtime > now && (((mtime % 2) == 0) && ((mtime-1) > now)))
+#else
+        if (mtime > now)
 #endif
-	  {
-	    error ("*** File `%s' has modification time in the future",
-		   file->name);
-	    file->update_status = 2;
-	  }
+          {
+            error("*** Warning: File `%s' has modification time in the future",
+                  file->name);
+            clock_skew_detected = 1;
+          }
       }
   }
 
@@ -1073,11 +1127,7 @@ name_mtime (name)
 {
   struct stat st;
 
-#ifdef VMS
-  if (vms_stat (name, &st) < 0)
-#else
   if (stat (name, &st) < 0)
-#endif
     return (time_t) -1;
 
   return (time_t) st.st_mtime;
@@ -1099,7 +1149,7 @@ library_search (lib, mtime_ptr)
       "/lib",
       "/usr/lib",
 #endif
-#if defined(WIN32) && !defined(LIBDIR)
+#if defined(WINDOWS32) && !defined(LIBDIR)
 /*
  * This is completely up to the user at product install time. Just define
  * a placeholder.
@@ -1145,6 +1195,30 @@ library_search (lib, mtime_ptr)
 
   /* Now try the standard set of directories.  */
 
+#ifdef  __MSDOS__
+  {
+    /* The default library directory is at ${DJDIR}/lib.  */
+    struct variable *djdir = lookup_variable ("DJDIR", 5);
+
+    if (djdir)
+      {
+	size_t djdir_len = strlen (djdir->value);
+
+	if (djdir_len > sizeof(LIBDIR) + 8 + strlen(libname) + 4 + 2)
+	  buf = (char *) xrealloc (djdir_len + 1);
+	sprintf (buf, "%s/lib/lib%s.a", djdir->value, libname);
+	mtime = name_mtime (buf);
+	if (mtime != (time_t) -1)
+	  {
+	    *lib = buf;
+	    if (mtime_ptr != 0)
+	      *mtime_ptr = mtime;
+	    return 1;
+	  }
+      }
+  }
+#endif
+
   for (dp = dirs; *dp != 0; ++dp)
     {
 #ifndef _AMIGA
