@@ -17,13 +17,14 @@ along with GNU Make; see the file COPYING.  If not, write to
 the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  */
 
 #include "make.h"
-#include "commands.h"
 #include "dep.h"
-#include "file.h"
+#include "filedef.h"
+#include "job.h"
+#include "commands.h"
 #include "variable.h"
 #include "rule.h"
 
-static void freerule ();
+static void freerule PARAMS ((struct rule *rule, struct rule *lastrule));
 
 /* Chain of all pattern rules.  */
 
@@ -96,9 +97,13 @@ count_implicit_rule_limits ()
       for (dep = rule->deps; dep != 0; dep = dep->next)
 	{
 	  unsigned int len = strlen (dep->name);
-	  char *p = rindex (dep->name, '/');
-	  char *p2 = p != 0 ? index (dep->name, '%') : 0;
 
+#ifdef VMS
+	  char *p = rindex (dep->name, ']');
+#else
+	  char *p = rindex (dep->name, '/');
+#endif
+	  char *p2 = p != 0 ? index (dep->name, '%') : 0;
 	  ndeps++;
 
 	  if (len > max_pattern_dep_length)
@@ -125,7 +130,11 @@ count_implicit_rule_limits ()
 		 nonexistent subdirectory.  */
 
 	      dep->changed = !dir_file_exists_p (name, "");
+#ifdef VMS
+	      if (dep->changed && *name == ']')
+#else
 	      if (dep->changed && *name == '/')
+#endif
 		{
 		  /* The name is absolute and the directory does not exist.
 		     This rule can never possibly match, since this dependency
@@ -164,7 +173,7 @@ convert_suffix_rule (target, source, cmds)
      char *target, *source;
      struct commands *cmds;
 {
-  char *targname, *depname;
+  char *targname, *targpercent, *depname;
   char **names, **percents;
   struct dep *deps;
   unsigned int len;
@@ -172,7 +181,14 @@ convert_suffix_rule (target, source, cmds)
   if (target == 0)
     /* Special case: TARGET being nil means we are defining a
        `.X.a' suffix rule; the target pattern is always `(%.o)'.  */
-    targname = savestring ("(%.o)", 5);
+    {
+#ifdef VMS
+      targname = savestring ("(%.obj)", 7);
+#else
+      targname = savestring ("(%.o)", 5);
+#endif
+      targpercent = targname + 1;
+    }
   else
     {
       /* Construct the target name.  */
@@ -180,11 +196,13 @@ convert_suffix_rule (target, source, cmds)
       targname = xmalloc (1 + len + 1);
       targname[0] = '%';
       bcopy (target, targname + 1, len + 1);
+      targpercent = targname;
     }
 
   names = (char **) xmalloc (2 * sizeof (char *));
   percents = (char **) alloca (2 * sizeof (char *));
-  names[0] = percents[0] = targname;
+  names[0] = targname;
+  percents[0] = targpercent;
   names[1] = percents[1] = 0;
 
   if (source == 0)
@@ -503,13 +521,37 @@ create_pattern_rule (targets, target_percents,
 
 /* Print the data base of rules.  */
 
+static void			/* Useful to call from gdb.  */
+print_rule (r)
+     struct rule *r;
+{
+  register unsigned int i;
+  register struct dep *d;
+
+  for (i = 0; r->targets[i] != 0; ++i)
+    {
+      fputs (r->targets[i], stdout);
+      if (r->targets[i + 1] != 0)
+	putchar (' ');
+      else
+	putchar (':');
+    }
+  if (r->terminal)
+    putchar (':');
+
+  for (d = r->deps; d != 0; d = d->next)
+    printf (" %s", dep_name (d));
+  putchar ('\n');
+
+  if (r->cmds != 0)
+    print_commands (r->cmds);
+}
+
 void
 print_rule_data_base ()
 {
   register unsigned int rules, terminal;
   register struct rule *r;
-  register struct dep *d;
-  register unsigned int i;
 
   puts ("\n# Implicit Rules");
 
@@ -519,26 +561,10 @@ print_rule_data_base ()
       ++rules;
 
       putchar ('\n');
-      for (i = 0; r->targets[i] != 0; ++i)
-	{
-	  fputs (r->targets[i], stdout);
-	  if (r->targets[i + 1] != 0)
-	    putchar (' ');
-	  else
-	    putchar (':');
-	}
+      print_rule (r);
+
       if (r->terminal)
-	{
-	  ++terminal;
-	  putchar (':');
-	}
-
-      for (d = r->deps; d != 0; d = d->next)
-	printf (" %s", dep_name (d));
-      putchar ('\n');
-
-      if (r->cmds != 0)
-	print_commands (r->cmds);
+	++terminal;
     }
 
   if (rules == 0)
@@ -548,6 +574,11 @@ print_rule_data_base ()
       printf ("\n# %u implicit rules, %u", rules, terminal);
 #ifndef	NO_FLOAT
       printf (" (%.1f%%)", (double) terminal / (double) rules * 100.0);
+#else
+      {
+	int f = (terminal * 1000 + 5) / rules;
+	printf (" (%d.%d%%)", f/10, f%10);
+      }
 #endif
       puts (" terminal.");
     }
