@@ -1,5 +1,5 @@
 /* Implementation of pattern-matching file search paths for GNU Make.
-Copyright (C) 1988, 89, 91, 92, 93, 94, 95, 96 Free Software Foundation, Inc.
+Copyright (C) 1988,89,91,92,93,94,95,96,97 Free Software Foundation, Inc.
 This file is part of GNU Make.
 
 GNU Make is free software; you can redistribute it and/or modify
@@ -19,7 +19,7 @@ the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  */
 #include "make.h"
 #include "filedef.h"
 #include "variable.h"
-#ifdef WIN32
+#ifdef WINDOWS32
 #include "pathstuff.h"
 #endif
 
@@ -43,6 +43,10 @@ static struct vpath *vpaths;
 /* Structure for the general VPATH given in the variable.  */
 
 static struct vpath *general_vpath;
+
+/* Structure for GPATH given in the variable.  */
+
+static struct vpath *gpaths;
 
 static int selective_vpath_search PARAMS ((struct vpath *path, char **file, time_t *mtime_ptr));
 
@@ -98,6 +102,38 @@ build_vpath_lists ()
       general_vpath = vpaths;
       vpaths = save_vpaths;
     }
+
+  /* If there is a GPATH variable with a nonnull value, construct the
+     GPATH list from it.  We use variable_expand rather than just
+     calling lookup_variable so that it will be recursively expanded.  */
+
+  {
+    /* Turn off --warn-undefined-variables while we expand SHELL and IFS.  */
+    int save = warn_undefined_variables_flag;
+    warn_undefined_variables_flag = 0;
+
+    p = variable_expand ("$(strip $(GPATH))");
+
+    warn_undefined_variables_flag = save;
+  }
+
+  if (*p != '\0')
+    {
+      /* Save the list of vpaths.  */
+      struct vpath *save_vpaths = vpaths;
+
+      /* Empty `vpaths' so the new one will have no next, and `vpaths'
+	 will still be nil if P contains no existing directories.  */
+      vpaths = 0;
+
+      /* Parse P.  */
+      construct_vpath_list ("%", p);
+
+      /* Store the created path as the GPATH,
+	 and restore the old list of vpaths.  */
+      gpaths = vpaths;
+      vpaths = save_vpaths;
+    }
 }
 
 /* Construct the VPATH listing for the pattern and searchpath given.
@@ -109,14 +145,15 @@ build_vpath_lists ()
    variable.
 
    If SEARCHPATH is nil, remove all previous listings with the same
-   pattern.  If PATTERN is nil, remove all VPATH listings.
-   Existing and readable directories that are not "." given in the
-   searchpath separated by colons are loaded into the directory hash
-   table if they are not there already and put in the VPATH searchpath
-   for the given pattern with trailing slashes stripped off if present
-   (and if the directory is not the root, "/").
-   The length of the longest entry in the list is put in the structure as well.
-   The new entry will be at the head of the VPATHS chain.  */
+   pattern.  If PATTERN is nil, remove all VPATH listings.  Existing
+   and readable directories that are not "." given in the searchpath
+   separated by the path element separator (defined in make.h) are
+   loaded into the directory hash table if they are not there already
+   and put in the VPATH searchpath for the given pattern with trailing
+   slashes stripped off if present (and if the directory is not the
+   root, "/").  The length of the longest entry in the list is put in
+   the structure as well.  The new entry will be at the head of the
+   VPATHS chain.  */
 
 void
 construct_vpath_list (pattern, dirpath)
@@ -173,14 +210,14 @@ construct_vpath_list (pattern, dirpath)
       return;
     }
 
-#ifdef WIN32
-    convert_vpath_to_win32(dirpath, ';');
+#ifdef WINDOWS32
+    convert_vpath_to_windows32(dirpath, ';');
 #endif
 
-  /* Figure out the maximum number of VPATH entries and
-     put it in MAXELEM.  We start with 2, one before the
-     first colon and one nil, the list terminator and
-     increment our estimated number for each colon or blank we find.  */
+  /* Figure out the maximum number of VPATH entries and put it in
+     MAXELEM.  We start with 2, one before the first separator and one
+     nil (the list terminator) and increment our estimated number for
+     each separator or blank we find.  */
   maxelem = 2;
   p = dirpath;
   while (*p != '\0')
@@ -190,7 +227,7 @@ construct_vpath_list (pattern, dirpath)
   vpath = (char **) xmalloc (maxelem * sizeof (char *));
   maxvpath = 0;
 
-  /* Skip over any initial colons and blanks.  */
+  /* Skip over any initial separators and blanks.  */
   p = dirpath;
   while (*p == PATH_SEPARATOR_CHAR || isblank (*p))
     ++p;
@@ -209,6 +246,10 @@ construct_vpath_list (pattern, dirpath)
       len = p - v;
       /* Make sure there's no trailing slash,
 	 but still allow "/" as a directory.  */
+#ifdef __MSDOS__
+      /* We need also to leave alone a trailing slash in "d:/".  */
+      if (len > 3 || (len > 1 && v[1] != ':'))
+#endif
       if (len > 1 && p[-1] == '/')
 	--len;
 
@@ -231,7 +272,7 @@ construct_vpath_list (pattern, dirpath)
 	    free (v);
 	}
 
-      /* Skip over colons and blanks between entries.  */
+      /* Skip over separators and blanks between entries.  */
       while (*p == PATH_SEPARATOR_CHAR || isblank (*p))
 	++p;
     }
@@ -270,6 +311,24 @@ construct_vpath_list (pattern, dirpath)
     }
 }
 
+/* Search the GPATH list for a pathname string that matches the one passed
+   in.  If it is found, return 1.  Otherwise we return 0.  */
+
+int
+gpath_search (file, len)
+     char *file;
+     int len;
+{
+  register char **gp;
+
+  if (gpaths && (len <= gpaths->maxlen))
+    for (gp = gpaths->searchpath; *gp != NULL; ++gp)
+      if (!strncmp(*gp, file, len) && (*gp)[len] == '\0')
+        return 1;
+
+  return 0;
+}
+
 /* Search the VPATH list whose pattern matches *FILE for a directory
    where the name pointed to by FILE exists.  If it is found, we set *FILE to
    the newly malloc'd name of the existing file, *MTIME_PTR (if MTIME_PTR is
@@ -287,8 +346,8 @@ vpath_search (file, mtime_ptr)
      there is nothing we can do.  */
 
   if (**file == '/'
-#ifdef WIN32
-      || **file == '\\' 
+#if defined (WINDOWS32) || defined (__MSDOS__)
+      || **file == '\\'
       || (*file)[1] == ':'
 #endif
       || (vpaths == 0 && general_vpath == 0))
@@ -343,9 +402,13 @@ selective_vpath_search (path, file, mtime_ptr)
      pointer to the name-within-directory and FLEN is its length.  */
 
   n = rindex (*file, '/');
-#ifdef WIN32
-  if (!n)
-    n = rindex(*file,, '\\');
+#if defined (WINDOWS32) || defined (__MSDOS__)
+  /* We need the rightmost slash or backslash.  */
+  {
+    char *bslash = rindex(*file, '\\');
+    if (!n || bslash > n)
+      n = bslash;
+  }
 #endif
   name_dplen = n != 0 ? n - *file : 0;
   filename = name_dplen > 0 ? n + 1 : *file;
@@ -373,18 +436,27 @@ selective_vpath_search (path, file, mtime_ptr)
       /* Add the directory prefix already in *FILE.  */
       if (name_dplen > 0)
 	{
+#ifndef VMS
 	  *n++ = '/';
+#endif
 	  bcopy (*file, n, name_dplen);
 	  n += name_dplen;
 	}
 
+#if defined (WINDOWS32) || defined (__MSDOS__)
+      /* Cause the next if to treat backslash and slash alike.  */
+      if (n != name && n[-1] == '\\' )
+	n[-1] = '/';
+#endif
       /* Now add the name-within-directory at the end of NAME.  */
+#ifndef VMS
       if (n != name && n[-1] != '/')
 	{
 	  *n = '/';
 	  bcopy (filename, n + 1, flen + 1);
 	}
       else
+#endif
 	bcopy (filename, n, flen + 1);
 
       /* Check if the file is mentioned in a makefile.  If *FILE is not
@@ -397,7 +469,7 @@ selective_vpath_search (path, file, mtime_ptr)
 	 inadequately commented change in July 1990; I am not sure off
 	 hand what problem it fixes.
 
-	 In December 1993 I loosened of this restriction to allow a file
+	 In December 1993 I loosened this restriction to allow a file
 	 to be chosen if it is mentioned as a target in a makefile.  This
 	 seem logical.  */
       {
@@ -411,6 +483,9 @@ selective_vpath_search (path, file, mtime_ptr)
 	  /* That file wasn't mentioned in the makefile.
 	     See if it actually exists.  */
 
+#ifdef VMS
+	  exists_in_cache = exists = dir_file_exists_p (vpath[i], filename);
+#else
 	  /* Clobber a null into the name at the last slash.
 	     Now NAME is the name of the directory to look in.  */
 	  *n = '\0';
@@ -419,6 +494,7 @@ selective_vpath_search (path, file, mtime_ptr)
 	     construct_vpath_list or the code just above put it there.
 	     Does the file we seek exist in it?  */
 	  exists_in_cache = exists = dir_file_exists_p (name, filename);
+#endif
 	}
 
       if (exists)
@@ -431,8 +507,10 @@ selective_vpath_search (path, file, mtime_ptr)
 
 	  struct stat st;
 
+#ifndef VMS
 	  /* Put the slash back in NAME.  */
 	  *n = '/';
+#endif
 
 	  if (!exists_in_cache	/* Makefile-mentioned file need not exist.  */
 	      || stat (name, &st) == 0) /* Does it really exist?  */
